local LibStub = LibStub
local D = LibStub:NewLibrary("DoomCore-2.0",3)
if not D then return end
local A = LibStub("Abacus-1.0")
local SharedMedia = LibStub( "LibSharedMedia-3.0" )
local AceAddon = LibStub("AceAddon-3.0")

function AA(addon)
  return AceAddon:GetAddon(addon, true)
end

local ceil, floor, max, min, next, rawget, pairs, pow, select, strbyte, tonumber, tostring, type, unpack
    = ceil, floor, max, min, next, rawget, pairs, pow, select, strbyte, tonumber, tostring, type, unpack
local error, print, setmetatable, tinsert
    = error, print, setmetatable, tinsert
local BNGetInfo, CreateFrame, InCombatLockdown, UIFrame, UIParent
    = BNGetInfo, CreateFrame, InCombatLockdown, UIFrame, UIParent
local bounded,   capitalize,   npairs,   sublist,   tappend,   tooltip,   tostrings
  = A.bounded, A.capitalize, A.npairs, A.sublist, A.tappend, A.tooltip, A.tostrings
local assertType,   colPack,   colUnpack,   Inherits,   setIndex,   scopify
  = A.assertType, A.colPack, A.colUnpack, A.Inherits, A.setIndex, A.scopify
-----------
-- Utility
-----------

local function orderNum(text)
  text = text:upper()
  local order = 0
  for i=1,#text do
    order = order + (strbyte(text, i)) / pow(100, i)
  end
  return order
end
D.orderNum = orderNum

-----------
-- Anchors
-----------

local anchors = {
  center      = "Center",
  topleft     = "Top Left",
  top         = "Top",
  topright    = "Top Right",
  left        = "Left",
  right       = "Right",
  bottomleft  = "Bottom Left",
  bottom      = "Bottom",
  bottomright = "Bottom Right"
}
D.anchors = anchors

local VERTICAL = 1
D.VERTICAL = VERTICAL
local HORIZONTAL = 2
D.HORIZONTAL = HORIZONTAL
local orientations = {
  horizontal = "Horizontal",
  vertical = "Vertical"
}
D.orientations = orientations
local orientGrowth = {
  horizontal = {
    [false] = "right",
    [true] = "left"
  },
  vertical = {
    [false] = "top",
    [true] = "bottom"
  }
}
D.orientGrowth = orientGrowth
local direction = {
  TOP    = VERTICAL,    top    = VERTICAL,
  BOTTOM = VERTICAL,    bottom = VERTICAL,
  LEFT   = HORIZONTAL,  left   = HORIZONTAL,
  RIGHT  = HORIZONTAL,  right  = HORIZONTAL,
  horizontal = HORIZONTAL,
  vertical = VERTICAL
}
D.direction = direction
local function compound(anchor1, anchor2)
  local dir1, dir2 = direction[anchor1], direction[anchor2]
  if not dir1 then return anchor1 end
  if not dir2 then return anchor2 end
  if dir1 < dir2 then
    return anchor1 .. anchor2
  else
    return anchor2 .. anchor1
  end
end
D.compound = compound

local grow = { left = "Left", right = "Right", top = "Up", bottom = "Down" }
D.grow = grow

local growAnchors = {
  left   = { "RIGHT",  "LEFT",  -1,   0 },
  right  = { "LEFT",   "RIGHT",  1,   0 },
  top    = { "BOTTOM", "TOP",    0,   1 },
  bottom = { "TOP",    "BOTTOM", 0,  -1 },
  blank  = { "CENTER", "CENTER", 0,   0 },
  center = { "CENTER", "CENTER", 0,   0 }
}
D.growAnchors = growAnchors
for _,dir1 in pairs{ "left", "right" } do for _,dir2 in pairs{ "top", "bottom" } do
  growAnchors[compound(growAnchors[dir1][1],growAnchors[dir2][1])] = { nil, nil, growAnchors[dir1][3], growAnchors[dir2][4] }
end end

local corners = {
  topleft     = { "BOTTOMRIGHT", "TOPLEFT",     -1, -1 },
  topright    = { "BOTTOMLEFT",  "TOPRIGHT",     1, -1 },
  bottomleft  = { "TOPRIGHT",    "BOTTOMLEFT",  -1,  1 },
  bottomright = { "TOPLEFT",     "BOTTOMRIGHT",  1,  1 }
}
D.corners = corners

local borders = {
  top    = { { "BOTTOMLEFT", "TOPLEFT",    -1,  0 }, { "BOTTOMRIGHT", "TOPRIGHT",     1,  0 },  0,  1 },
  bottom = { { "TOPLEFT",    "BOTTOMLEFT", -1,  0 }, { "TOPRIGHT",    "BOTTOMRIGHT",  1,  0 },  0,  1 },
  left   = { { "TOPRIGHT",   "TOPLEFT",     0,  1 }, { "BOTTOMRIGHT", "BOTTOMLEFT",   0, -1 },  1,  0 },
  right  = { { "TOPLEFT",    "TOPRIGHT",    0,  1 }, { "BOTTOMLEFT",  "BOTTOMRIGHT",  0, -1 },  1,  0 }
}
D.borders = borders

local function convertDims(dir, dim1, dim2)
  if direction[dir] == HORIZONTAL then
    return dim1, dim2
  else
    return dim2, dim1
  end
end
D.convertDims = convertDims


----------
-- Frames
----------

local function Locker()
  local locker = DoomCoreLocker
  if locker then return locker end
  locker = CreateFrame("Button", "DoomCoreLocker", UIParent, "UIPanelButtonTemplate")
  locker:SetFrameStrata("DIALOG")
  locker.frames = {}
  locker:SetText("Lock all")
  locker:SetSize(80, 25)
  locker:SetPoint("BOTTOM", 0, 150)
  locker:Hide()
  function locker:Update()
    local visible = false
    for frameName, frame in pairs(self.frames) do
      if frame and frame.conf and frame.conf.lock == false then
        visible = true
        break
      end
    end
    if visible then self:Show() else self:Hide() end
  end
  locker:SetScript("OnClick", function(self)
    for frameName, frame in pairs(self.frames) do
      if frame.conf then frame.conf.lock = true end
      if frame.tex then frame.tex:Hide() end
    end
    self.frames = {}
    self:Hide()
  end)
  return locker
end
D.Locker = Locker


local function pad(frame, padding, parent)
  parent = parent or frame:GetParent()
  frame:ClearAllPoints()
  for _,corner in pairs(corners) do
    frame:SetPoint(corner[1], parent, corner[1], padding * corner[3], padding * corner[4])
  end
end
D.pad = pad
local function updateFrame(frame, c, anchor, parent)
  c = c or frame.conf or {}
  anchor = anchor or c.anchor
  parent = parent or UIFrame

  local height,   lock,   offsetX,   offsetY,   scale,   width
    = c.height, c.lock, c.offsetX, c.offsetY, c.scale, c.width
  if anchor and not anchors[anchor:lower()] then anchor = nil end

  local unlocked = false
  if width then frame:SetWidth(width) end
  if height then frame:SetHeight(height) end
  if anchor then
    frame:ClearAllPoints()
    frame:SetPoint(anchor, parent, anchor, offsetX or 0, offsetY or 0)
  end
  if lock ~= nil then
    unlocked = not lock
    local locker = Locker()
    locker.frames[frame:GetName()] = frame
    locker:Update()
  end

  if scale then frame:SetScale(scale) end

  frame:SetMovable(unlocked)
  frame:EnableMouse(unlocked)
  if not frame.tex then
  elseif unlocked then
    frame.tex:Show()
  else
    frame.tex:Hide()
  end
  for _,child in pairs{ frame:GetChildren() } do
    if child.EnableMouse then child:EnableMouse(not unlocked) end
  end
  return unlocked
end
D.updateFrame = updateFrame

local function draggable(frame, frameSettings, func)
  if frame:GetScript("OnReceiveDrag") then return end
  frame:RegisterForDrag("LeftButton")
  frame:SetScript("OnDragStart", frame.StartMoving)
  frame:SetScript("OnDragStop", frame.StopMovingOrSizing)
  frame:SetScript("OnReceiveDrag", function()
    local orig,_,dest,x,y = frame:GetPoint(1)
    local a = growAnchors[orig]
    if frameSettings then
      frameSettings.anchor = dest:lower()
      frameSettings.offsetX = floor(x*10)/10
      frameSettings.offsetY =  floor(y*10)/10
    end
    if func then func() end
  end)
end
D.draggable = draggable

local function setFont(text, conf)
  local font = SharedMedia:Fetch(SharedMedia.MediaType.FONT, conf.font)
  text:SetFont(font, conf.fontSize)
  if conf.fontColor then text:SetTextColor(colUnpack(conf.fontColor)) end
end
D.setFont = setFont
local function setBg(frame, background, backgroundColor, inset, border, borderColor, edge)
  edge = edge or 4
  inset = inset or 4
  backgroundColor = backgroundColor or { r=1,g=1,b=1,a=1 }
  borderColor = borderColor or { r=1,g=1,b=1,a=1 }
  local bg = {}
  if background and background ~= "None" then
    bg.bgFile = SharedMedia:Fetch(SharedMedia.MediaType.BACKGROUND, background)
  end
  if border and border ~= "None" then
    bg.edgeFile = SharedMedia:Fetch(SharedMedia.MediaType.BORDER, border)
    bg.edgeSize = edge
  end
  bg.insets = { left = inset, right = inset, top = inset, bottom = inset }
  frame:SetBackdrop(bg)
  frame:SetBackdropColor(colUnpack(backgroundColor))
  frame:SetBackdropBorderColor(colUnpack(borderColor))
end
D.setBg = setBg
local function setBgFromConf(frame, conf)
  conf = conf or frame.conf
  setBg(frame, conf.background, conf.backgroundColor, conf.inset, conf.border, conf.borderColor, conf.edge)
end
D.setBgFromConf = setBgFromConf

local function makeGrid(c)
  c.padding = c.padding or 0
  c.spacing = c.spacing or 0

  if not c.grow2 then
    if direction[c.grow] == HORIZONTAL then
      c.grow2 = c.rowGrowth
    else
      c.grow2 = c.columnGrowth
    end
  end

  local anchor,   frame,   grow,   grow2,   parent,   x,   y
  = c.anchor, c.frame, c.grow, c.grow2, c.parent, c.x, c.y
  local els,   limit,  cMax,   padding,   size,   spacing
  = c.els, c.limit, c.max, c.padding, c.size, c.spacing

  local a1 = growAnchors[grow] or growAnchors.blank
  local a2 = growAnchors[grow2] or growAnchors.blank
  local comp = compound(a1[1], a2[1])
  local growComp = growAnchors[comp]

  frame:ClearAllPoints()
  frame:SetPoint(comp, parent, anchor,
    --conf_x - (conf_size + conf_padding * 2) * growComp[3] / 2,
    --conf_y - (conf_size + conf_padding * 2) * growComp[4] / 2)
    x, y)

  if not els or #els == 0 then
    frame:SetSize(size + padding * 2, size + padding * 2)
    return
  end

  els[1]:ClearAllPoints()
  els[1]:SetPoint(comp, frame, comp, growComp[3] * padding, growComp[4] * padding)
  els[1]:Show()

  local toMax = #els
  if cMax and cMax > 0 then toMax = min(toMax, cMax) end

  for i=2,toMax do
    els[i]:ClearAllPoints()

    if limit and limit > 0 and (i-1) % limit == 0 then
      els[i]:SetPoint(a2[1], els[i-limit], a2[2], a2[3] * spacing, a2[4] * spacing)
    else
      els[i]:SetPoint(a1[1], els[i-1], a1[2], a1[3] * spacing, a1[4] * spacing)
    end
    els[i]:Show()
  end
  for i=toMax+1,#els do
    els[i]:Hide()
  end
  local dim1, dim2
  if limit and limit > 0 and toMax > limit then
    dim1 = padding * 2 + size * limit + spacing * (limit - 1)
    local subs = ceil(toMax/limit)
    dim2 = padding * 2 + size * subs + spacing * (subs - 1)
  else
    dim1 = padding * 2 + size * toMax + spacing * (toMax - 1)
    dim2 = padding * 2 + size
  end
  frame:SetSize(convertDims(grow, dim1, dim2))
  setBgFromConf(frame, c)
  --conf_frame:Show()
end
D.makeGrid = makeGrid

local ScrollGrid = function(name, secure, conf, parent, masque)
  parent = parent or UIParent
  local inherit = "ActionButtonTemplate"
  if secure then inherit = inherit .. ",SecureActionButtonTemplate" end
  local frame = CreateFrame("Frame", name, parent)
  local slider = CreateFrame("Slider", name .. ".Slider", frame)
  slider:SetPoint("TOPLEFT", frame, "TOPRIGHT", -6, -10)
  slider:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", -6, 12)
  slider:SetThumbTexture("Interface\\DialogFrame\\UI-DialogBox-Background-Dark")
  slider:GetThumbTexture():SetSize(5,15)
  slider:SetWidth(5)
  frame.Slider = slider
  frame:EnableMouseWheel(true)
  frame:SetScript("OnMouseWheel", function(self, delta)
    self.i = bounded(0, (self.i or 0) + self.conf.cols * delta * -1, self.maxI or 0)
    self.Slider:SetValue(self.i)
    self:Display()
  end)

  if masque then
    frame.msq = masque:Group(name)
  end
  frame.conf = conf
  local inner = CreateFrame("Frame", name .. ".Inner", frame)
  inner:SetAllPoints()
  inner:Show()
  frame.Inner = inner

  function frame:FindMaxI()
    self.maxI = 0
    local els, rows, cols = self.els, self.conf.rows, self.conf.cols
    if not els then return end
    local elsSize = #els
    if elsSize == 0 then return end
    local remainder = elsSize % cols
    local full = elsSize
    if remainder > 0 then full = full + (cols - remainder) end
    self.maxI = max(0, full - cols * rows)
    self.Slider:SetMinMaxValues(0, self.maxI)
    self.Slider:SetValue(self.i)
  end

  function frame:Refresh(show)
    if InCombatLockdown() then return print("Can't modify",self:GetName(),"in combat.") end
    local conf,inner = self.conf, self.Inner
    local padding, cols, rows = conf.padding, conf.cols, conf.rows
    local iconSize, spacing = conf.iconSize, conf.spacing
    local charmSize = conf.charmSize or 14
    self:SetSize(
      padding * 2 + cols * iconSize + (cols - 1) * spacing,
      padding * 2 + rows * iconSize + (rows - 1) * spacing
    )
    local displayCount = cols * rows
    local children = { inner:GetChildren() }
    local childrenCount = #children

    local display = {}

    for i = 1, displayCount do
      local child = children[i]
      if not child then
        child = CreateFrame("Button", self:GetName() .. i, inner, inherit)
        child.Charms = {}
        child.Overlays = {}
        if self.msq then self.msq:AddButton(child) end
        children[i] = child
      end
      for _,v in pairs(child.Charms) do v:SetSize(charmSize, charmSize) end
      for _,v in pairs(child.Overlays) do v:SetSize(charmSize, charmSize) end
      tooltip(child, conf.tooltip, conf.tooltipOverride and conf.tooltipAnchor)
      child:SetSize(iconSize, iconSize)
      tinsert(display, child)
    end
    for i = displayCount + 1, childrenCount do
      children[i]:Hide()
    end

    makeGrid{
      frame        = inner,
      els          = display,
      parent       = self,
      size         = iconSize,
      padding      = padding,
      spacing      = spacing,
      limit        = cols,
      grow         = "right",
      grow2        = "bottom",
    }
    self.display = display

    updateFrame(self, conf)
    setBgFromConf(self, conf)
    self:FindMaxI()
    local slider = self.Slider
    self:Display(show)
  end

  function frame:SetEls(els)
    self.els = els
    self.i = 0
    self:FindMaxI()
  end

  function frame:GetButtons()
    return self.Inner:GetChildren()
  end

  function frame:Display(show)
    if InCombatLockdown() then
      if show == false then
        print("Can't hide",self:GetName(),"in combat.")
      else
        print("Can't display",self:GetName(),"in combat.")
      end
      return
    end
    if show == false then return self:Hide() end
    local i, display, els, conf = self.i or 0, self.display, self.els, self.conf
    local charmSize = conf.charmSize or 14
    if not els then return end
    for childI,child in pairs(display) do
      local el = els[i + childI]
      if not el then
        child:Hide()
      else
        for _,charm in pairs(child.Charms) do charm:Hide() end
        for _,charm in pairs(child.Overlays) do charm:Hide() end
        --child:SetAttribute("unit","player")
        for k,v in pairs(el) do
          if k == "icon" then
            child.icon:SetTexture(v)
          elseif k == "tooltip" then
            child.tooltip = v
          elseif k == "item" or k == "spell" or k == "macro" then
            child:SetAttribute("type", k)
            if k == "macro" then k = "macrotext" end
            child:SetAttribute(k, v)
          elseif k == "charms" and v then
            local i = 1
            local charms = child.Charms
            for _,tex in pairs(v) do
              if tex then
                local charm = charms[i]
                if not charm then
                  charm = child:CreateTexture(child:GetName() .. ".Charms[" .. i .. "]", "OVERLAY")
                  charm:SetDrawLayer("OVERLAY", min(i, 7))
                  if i == 1 then
                    charm:SetPoint("TOPLEFT", child, "TOPLEFT", -1, 3)
                  else
                    charm:SetPoint("LEFT", charms[i-1], "RIGHT", 0.5, 0)
                  end
                  charm:SetSize(charmSize, charmSize)
                  charms[i] = charm
                end
                if tex == "star" then
                  charm:SetTexture(922035)
                  charm:SetTexCoord(0.13,0.72,0.08,0.67)
                else
                  charm:SetTexture(tex)
                  charm:SetTexCoord(0,1,0,1)
                end
                charm:Show()
                i = i + 1
              end
            end
          elseif k == "overlays" and v then
            local i = 1
            local overlays = child.Overlays
            for iLevel, over in pairs(v) do
              if over then
                local overlay = overlays[i]
                if not overlay then
                  overlay = child:CreateTexture(child:GetName() .. ".Overlays[" .. i .. "]", "OVERLAY")
                end
                local tex, color, anchor = over.tex, over.color, {unpack(over.anchor)}
                local size = over.size or charmSize
                tinsert(anchor, 2, child)
                overlay:ClearAllPoints()
                overlay:SetPoint(unpack(anchor))
                overlay:SetSize(size, size)
                overlays[i] = overlay
                if tex then
                  if tex == "star" then
                    overlay:SetTexture(922035)
                    overlay:SetTexCoord(0.13,0.72,0.08,0.67)
                  else
                    overlay:SetTexture(tex)
                    overlay:SetTexCoord(0,1,0,1)
                  end
                  if color then overlay:SetVertexColor(colUnpack(color))
                  else overlay:SetVertexColor(1,1,1)
                  end
                elseif color then
                  color.a = color.a or 1
                  overlay:SetColorTexture(colUnpack(color))
                end
                overlay:SetDrawLayer("OVERLAY",-7 + min(iLevel, 7))
                overlay:Show()
                i = i + 1
              end
            end
          else
            child:SetAttribute(k,v)
          end
        end
        child:Show()
      end
    end
    if self.msq then self.msq:ReSkin() end
    if show then self:Show() end
  end

  frame:Refresh()
  return frame
end
D.ScrollGrid = ScrollGrid

-------------------
-- Crawling
-------------------

local function crawl(obj, key, noChildren, raw)
  assertType(obj,"table", key,"table", 3)
  local max = #key
  if max == 0 then return obj elseif max < 0 then return nil end
  local node = obj
  local child
  for i=1,max do
    if raw then
      child = rawget(node, key[i])
    else
      child = node[key[i]]
    end
    if not noChildren and child == nil then
      child = {}
      node[key[i]] = child
    end
    node = child
    if node == nil then return end
  end
  return node
end
D.crawl = crawl

local NodeCrawler = Inherits:New()
function NodeCrawler:New(core)
  core = core or {}
  assertType(core,"table", 3)
  local new = NodeCrawler._super.New(self)
  new.core = core
  return new
end
function NodeCrawler:Get(key)
  assertType(key,"table", 3)
  return crawl(self.core, key, true)
end
function NodeCrawler:Set(key, value, noOverride)
  assertType(key,"table", 3)
  if not next(key) then
    if value == nil then value = {} end
    if not noOverride then self.core = value end
    return self.core
  end
  local parent = crawl(self.core, sublist(key, 1, -2), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local child = key[#key]
  if not (noOverride and parent[child] ~= nil) then
    parent[child] = value
  end
  return parent[child]
end
function NodeCrawler:GetOrMake(key)
  assertType(key,"table", 3)
  return self:Set(key, {}, true)
end
function NodeCrawler:Incr(key, amount)
  amount = amount or 1
  assertType(key,"table", amount,"number", 3)
  local parent = crawl(self.core, sublist(key, 1, -2), false, true)
  local child = key[#key]
  parent[child] = amount + (parent[child] or 0)
  return parent[child]
end
local function nappend(parent, key, ...)
  assertType(parent,"table", key,"string/number/boolean", 3)
  local child = parent[key]
  if child == nil then
    child = {}
    parent[key] = child
  elseif type(child) ~= "table" then
    return
  end
  tappend(child, ...)
end
D.nappend = nappend
function NodeCrawler:Append(key, ...)
  assertType(key,"table", 3)
  local parent = crawl(self.core, sublist(key, 1, -2), false, true)
  if parent ~= nil and type(parent) ~= "table" then return end
  local childKey = key[#key]
  nappend(parent, childKey, ...)
end
D.NodeCrawler = NodeCrawler

local Configuration = NodeCrawler:New()
function Configuration:New(val)
  local new = Configuration._super.New(self, val)
  return new
end
function Configuration:AutoDefault(child)
  if not child or not child.type or not self.defaults then return end
  local default = self.defaults[child.type]
  if default then setIndex(child, default, true) end
end
function Configuration:Inherit(child, parent)
  local defaults = self.defaults
  defaults[child] = defaults[child] or {}
  defaults[parent] = defaults[parent] or {}
  setIndex(defaults[child], defaults[parent], true)
end

function Configuration:CrawlDefaults(node, crawled)
  crawled = crawled or {}
  if not node then node = self.core end
  crawled[node] = true
  if type(node) ~= "table" then return end
  self:AutoDefault(node)
  for k,v in pairs(node) do
    if not crawled[v] then
      if type(v) == "table" and not (type(k) == "string" and k:find("_")) then
        self:CrawlDefaults(v, crawled)
      end
    end
  end
end
function Configuration:Make(parent, makeType, childName, noOverride)
  assertType(parent,"table", makeType,"string", childName,"string/number/boolean", 3)
  if type(childName) == "string" then
    childName = childName:gsub("_","")
    if childName == "" then return end
  end
  if noOverride and Configuration:Overriden(parent, childName) then return parent[childName] end
  local child = parent[childName]
  if child == nil then
    child = {}
    parent[childName] = child
  end
  child.type = makeType
  self:AutoDefault(child)
  return child
end
function Configuration:GetAll(parent, getType, depth)
  if not parent or not next(parent) then return pairs{} end
  depth = depth or 2
  assertType(parent,"table", getType,"string", depth,"number", 3)
  return npairs(parent, depth,
    function(val)
      return type(val) == "table" and val.type == getType
    end
  )
end
function Configuration:Count(parent, getType, depth)
  assertType(parent,"table", getType,"string", 3)
  local count = 0
  for ele in self:GetAll(parent, getType, depth) do
    count = count+1
  end
  return count
end
function Configuration:Delete(parent, childName)
  assertType(parent,"table", childName,"string/number/boolean", 3)
  parent[childName] = nil
  parent._exclude = parent._exclude or {}
  parent._exclude[childName] = true
end
function Configuration:Excluded(parent, childName)
  assertType(parent,"table", childName,"string/number/boolean", 3)
  if type(parent) ~= "table" then return end
  local exclude = parent._exclude
  return exclude and exclude[childName]
end
function Configuration:Overriden(parent, childName)
  assertType(parent,"table", childName,"string/number/boolean", 3)
  local child = parent[childName]
  if child == nil then return false end
  return child.type ~= nil
end
function Configuration:ConfGet(info)
  local val = self:Get(info)

  if type(val) == "table" and val.r and val.g and val.b then return colUnpack(val) end
  return val
end
function Configuration:ConfGetString(info)
  return tostring(self:ConfGet(info))
end
function Configuration:ConfSet(info, r, g, b, a)
  if b then return self:Set(info, colPack(r,g,b,a))
  elseif tonumber(r) then return self:Set(info, tonumber(r))
  else return self:Set(info, r) end
end
function Configuration:GetParent(info)
  local parent = self:Get(sublist(info, 1, -2))
  if type(parent) == "table" then return parent end
end
function Configuration:GetParentName(info)
  return info[#info - 1]
end
function Configuration:ConfGetAll(info)
  local parent = self:GetParent(info)
  if parent then return self:GetAll(parent, info[#info]) end
end
function Configuration:ConfDelete(info)
  local parent = self:GetParent(info)
  if parent then return self:Delete(parent, info[#info]) end
end
function Configuration:ConfExcluded(info)
  local parent = self:GetParent(info)
  if parent then return self:Excluded(parent, info[#info]) end
end
function Configuration:ConfOverriden(info)
  local parent = self:GetParent(info)
  if parent then return self:Overriden(parent, info[#info]) end
end
function Configuration:Number(info, val)
  return val and type(tonumber(val)) == "number"
end
function Configuration:Check(t, info)
  local size = #t
  if size == 0 then return true end

  local start = 1
  local sign
  if t[start] then
    sign = true
  else
    start = start + 1
    sign = false
  end

  if type(t[start + 1]) == "table" then
    local op = t[start]
    if op == "and" then
      for i=start+1, size do
        if not self:Check(t[i]) then return not sign end
      end
      return sign
    else
      for i=start+1, size do
        if self:Check(t[i]) then return sign end
      end
      return not sign
    end
  end

  local db = self.core
  local pref = t[start]
  if info and type(pref) == "string" then
    if pref:sub(1,1) == "^" then
      start = start + 1
      local parent = tonumber(pref:sub(2))
      if parent then
        local subinfo = {}
        for i=1,#info-parent do
          subinfo[i] = info[i]
        end
        db = self:Get(subinfo)
      else
        db = self:Get(info)
      end
    end
  end
  for i=start,size do
    if type(db) ~= "table" then return not sign end
    db = db[t[i]]
  end
  if db then return sign else return not sign end
end

D.Configuration = Configuration


-----------------
-- Addon handler
-----------------

local Handler = Configuration:New()

function Handler:New(addon)
  assertType(addon,"table", 3)
  local handler = Handler._super.New(self)
  setmetatable(addon, Handler)
  local addon_settings = addon.settings
  if addon_settings then
    addon.defaults = addon_settings.defaults
  end
  addon.timers = {}
  addon.frames = {}
  return addon
end
function D.Addon(shortName, name, N, ...)
  local addon = AceAddon:NewAddon(shortName, "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", ...)
  addon.shortName = shortName
  addon.name = name
  addon.mediaPath = "Interface\\AddOns\\" .. shortName .. "\\Media\\"
  if N then addon.settings = N.Settings end
  local handler = Handler:New(addon)
  if N then N.Handler = handler end
  return handler
end
function Handler:Output(...)
  if self.core._debug and self.core._debug.print then print(self.name," | ",tostrings(...)) end
end
function Handler:GetVersion()
  return self.core._version
end
function Handler:SetVersion(version)
  self.core._version = version
end
function Handler:ResetDB()
  self.lib.db:ResetProfile()
end
function Handler:Queue(funcName, ...)
  local timers = self.timers
  if type(self[funcName]) ~= "function" then error(tostrings("Invalid function ", funcName,"."), 2) end
  if not self.ScheduleTimer then error("AceTimer not mixed in!", 2) end
  if not timers[funcName] or self:TimeLeft(timers[funcName]) == 0 then
    timers[funcName] = self:ScheduleTimer(funcName,0.01,...)
  end
end
function Handler:TrackEvent(funcName, event, ...)
  event = event or funcName
  self:RegisterEvent(event, funcName, ...)
end
function Handler:TrackEventQueue(funcName, event, ...)
  event = event or funcName
  self:RegisterEvent(event, "Queue", funcName, ...)
end
function Handler:TrackMultiEvent(funcName, ...)
  self:Map("TrackEvent", { funcName }, {...})
end
function Handler:TrackMultiEventQueue(funcName, ...)
  self:Map("TrackEventQueue", { funcName }, {...})
end

function Handler:Reset(registered)
  self.core = self.lib.db.profile
  self:CrawlDefaults()
  if registered ~= false then
    self:SetVersion(self.version)
  end

  local settings = self.settings
  if settings then
    settings.options = { type = "group", args = {} }
    local defaults = self.defaults
    if defaults and defaults.options then
      settings.options = defaults.options
    end
    settings.options.handler = self
    settings.crawler = NodeCrawler:New(settings.options.args)

    if defaults and defaults.global then
      setIndex(self.core, defaults.global)
    end
  end
  if self.OnLoad then self:OnLoad(registered) end

  if settings and settings.options then
    self.lib.registry:RegisterOptionsTable(self.name, settings.options)
  end
end

function Handler:MigrateDB(...)
  local old = Configuration:New{}
  for i=1,select("#", ...) do
    local key = select(i, ...)
    if key == nil then key = {} end
    local subTable = self:Get(key)
    if type(subTable) ~= "table"then
      old:Set(key, subTable)
    else
      for k,v in pairs(subTable) do
        if type(v) ~= "table" then old:Set(tappend({unpack(key)},k), v) end
      end
    end
  end

  self:ResetDB()

  for i=1,select("#", ...) do
    local key = select(i, ...)
    local subTable = old:Get(key)
    if type(subTable) ~= "table" then
      if subTable ~= self:ConfGet(key) then self:ConfSet(key, subTable) end
    else for k,v in pairs(subTable) do
      local subKey = tappend({unpack(key)}, k)
      if type(v) ~= "table" and v ~= self:ConfGet(subKey) then self:ConfSet(subKey, v) end
    end end
  end

  if self.version then self:SetVersion(self.version) end
end

function Handler:ConfSetReload(info, r, g, b, a)
  self:ConfSet(info, r, g, b, a)
  ReloadUI()
end
local function debugKey(info)
  return tappend({ "_debug" }, unpack(info))
end
function Handler:ConfGetDebug(info)
  return self:ConfGet(debugKey(info))
end
function Handler:ConfSetDebug(info, r, g, b, a)
  self:ConfSet(debugKey(info), r, g, b, a)
end
function Handler:DebugOptions(args)
  local debug = {
    type = "group",
    handler = self,
    get = "ConfGetDebug",
    set = "ConfSetDebug",
    args = {
      export = {
        type = "description",
        name = 'Exported to AA("' .. self.shortName .. '")',
        width = "full",
        fontSize = "medium",
        order = 0
      },
      print = {
        type = "toggle",
        name = "Print debug messages",
        width = "full",
        set = "ConfSetDebug",
        order = 1
      },
    }
  }
  local debug_args = debug.args
  for k,v in pairs(args) do
    debug_args[k] = v
  end
  return debug
end

local dbCallbacks = { "OnProfileCopied", "OnProfileDeleted", "OnProfileReset", "OnDatabaseReset", "OnProfileChanged" }
function Handler:Register(libs,defaultProfile,defaults)
  self.lib = self.lib or {}
  local name, lib, settings = self.name, self.lib, self.settings
  if settings and settings.custom then
    settings.defaults = settings.defaults or {}
    local subDefaults = settings.defaults
    local battletag = select(2, BNGetInfo())
    for name, custom in pairs(settings.custom) do
      if name == battletag then
        for customType, customSettings in pairs(custom) do
          subDefaults[customType] = subDefaults[customType] or {}
          local subDefaultType = subDefaults[customType]
          for propertyName, propertyVal in pairs(customSettings) do
            subDefaultType[propertyName] = propertyVal
          end
        end
      end
    end
    if not defaults and subDefaults.db then
      defaults = settings.defaults.db
    end
  end
  defaults = defaults or {}

  lib.db = LibStub("AceDB-3.0"):New(self.shortName .. "DB", defaults, defaultProfile)

  lib.registry = LibStub("AceConfigRegistry-3.0")
  lib.settings = LibStub("AceConfigDialog-3.0")
  lib.options = LibStub("AceDBOptions-3.0")
  for libKey, libName in pairs(libs or {}) do
    if not lib[libKey] then
      lib[libKey] = LibStub(libName,true) or AceAddon:GetAddon(libName,true)
    end
  end

  self:Reset(false)

  if settings then
    lib.settings:AddToBlizOptions(name)
    local profileSettings = LibStub("AceDBOptions-3.0"):GetOptionsTable(lib.db)
    lib.registry:RegisterOptionsTable(name .. " Profile", profileSettings)
    lib.settings:AddToBlizOptions(name .. " Profile", "Profiles", name)

    if settings.debug then
      local debug = self:DebugOptions(settings.debug)
      lib.registry:RegisterOptionsTable(name .. " Extras", debug)
      lib.settings:AddToBlizOptions(name .. " Extras","Extras", name)
    end
  end

  for _,callback in pairs(dbCallbacks) do
    lib.db.RegisterCallback(self, callback, "Reset")
  end
end

function Handler:Draggable(frame, conf, funcName)
  if frame:GetScript("OnReceiveDrag") then return end
  funcName = funcName or "Rebuild"
  local func = function()
    if type(funcName) == "function" then funcName()
    elseif type(self[funcName]) == "function" then self[funcName](self) end
    self.lib.registry:NotifyChange(self.name)
  end
  draggable(frame, conf, func)
end

D.Handler = Handler


----------------------
-- Settings templates
----------------------

local optTs = {}
optTs.parent = {
  type = "group",
  get = "ConfGet",
  set = "ConfSet"
}
optTs.space = {
  type = "description",
  name = "",
  width = "full"
}
optTs.label = {
  type = "description",
  width = "half",
  fontSize = "medium"
}
optTs.reload = {
  type = "toggle",
  set = "ConfSetReload",
  desc = "Requires UI reload",
  confirm = function() return "Requires Reload UI" end
}
optTs.lock = {
  type = "toggle",
  name = "Lock"
}
optTs.background = {
  name = "Background",
  type = "select",
  dialogControl = "LSM30_Background",
  values = SharedMedia:HashTable( SharedMedia.MediaType.BACKGROUND )
}
optTs.border = {
  name = "Border",
  type = "select",
  dialogControl = "LSM30_Border",
  values = SharedMedia:HashTable( SharedMedia.MediaType.BORDER )
}
optTs.font = {
  name = "Font",
  type = "select",
  dialogControl = "LSM30_Font",
  values = SharedMedia:HashTable(SharedMedia.MediaType.FONT )
}
optTs.range = {
  type = "range",
  step = 0.5
}
optTs.fontSize = {
  type = "range",
  name = "Font size",
  min = 1,
  max = 100,
  softMin = 10,
  softMax = 50,
  step = 1
}
optTs.iconSize = {
  type = "range",
  name = "Icon size",
  min = 1,
  max = 100,
  softMin = 5,
  softMax = 50,
  step = 0.5
}
optTs.iconSpacing = {
  type = "range",
  name = "Icon Spacing",
  min = -100,
  max = 100,
  softMin = -20,
  softMax = 20,
  step = 0.5
}
optTs.bigIconSize = {
  type = "range",
  name = "Icon size",
  min = 1,
  max = 500,
  softMin = 10,
  softMax = 50,
  step = 0.5
}
optTs.edge = {
  type = "range",
  name = "Border size",
  min = 1,
  max = 100,
  softMax = 36,
  step = 0.5
}
optTs.inset = {
  type = "range",
  name = "Inset",
  min = 0,
  max = 100,
  softMax = 36,
  step = 0.5
}
optTs.anchor = {
  type = "select",
  name = "Anchor",
  values = anchors
}
optTs.offsetX = {
  type = "range",
  name = "X",
  min = -floor(UIParent:GetWidth()),
  max = floor(UIParent:GetWidth()),
  step = 0.5,
  disabled = ({ "^1", "lock" })
}
optTs.offsetY = {
  type = "range",
  name = "Y",
  min = -floor(UIParent:GetHeight()),
  max = floor(UIParent:GetHeight()),
  step = 0.5,
  disabled = ({ "^1", "lock" })
}
optTs.grow = {
  type = "select",
  name = "Grow",
  values = grow
}
optTs.orientation = {
  type = "select",
  name = "Orientation",
  values = D.orientations,
  order = 12
}
optTs.aColor = {
  name = "Color",
  type = "color",
  hasAlpha = true,
  width = "half"
}
optTs.percent = {
  type = "range",
  min = 0,
  max = 1,
  step = 0.01,
  isPercent = true
}
optTs.scale = {
  type = "range",
  name = "Scale",
  min = 0.001,
  max = 10,
  softMax = 2,
  step = 0.001,
  isPercent = true
}

local checks = {}
for _,check in pairs{ "disabled", "hidden", "confirm" } do checks[check] = true end
local function ensureCheck(k, v)
  if checks[k] and type(v) == "table" then return function(info)
    return info.handler:Check(v, info)
  end end
  return v
end

local function optCopy(optOut, body)
  for k,v in pairs(body or {}) do
    if k:sub(1,1) == "_" then
      optOut[k:sub(2)] = v
    elseif k == "values" then
      optOut[k] = optOut[k] or {}
      local values = optOut[k]
      for i,value in pairs(v) do
        if type(i) == "number" then
          values[value] = capitalize(value)
        else
          values[i] = value
        end
      end
    elseif k == "args" then
      optOut[k] = optOut[k] or {}
      local args = optOut[k]
      for i,arg in pairs(v) do
        if type(i) == "number" then
          local newI, optType, name, body = unpack(arg)
          if type(name) == "table" then
            body = name
            name = nil
          elseif type(optType) == "table" then
            body = optType
            optType = nil
            name = nil
          end
          body = body or {}
          if name then
            body.name = body.name or name
          else
            name = capitalize(newI)
          end
          optType = optType or newI
          if args[newI] then newI = newI .. i end
          args[newI] = D.opt(i, optType, name, body)
        else
          args[i] = arg
        end
      end
    else
      optOut[k] = ensureCheck(k, v)
    end
  end
end

local function opt(order, optType, name, body)
  local optOut = {}
  if order == "parent" then
    order = nil
    name = nil
    body = optType
    optType = "parent"
  elseif type(name) == "table" then
    body = name
    name = optType
  end
  for k,v in pairs(optTs[optType] or {}) do optOut[k] = ensureCheck(k, v) end
  optOut.type = optOut.type or optType
  optOut.name = optOut.name or name
  optOut.order = order
  optCopy(optOut, body)
  return optOut
end
D.opt = opt